#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google DisplayVideo operators."""

from __future__ import annotations

import csv
import json
import os
import shutil
import tempfile
import urllib.request
import zipfile
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any
from urllib.parse import urlsplit

from airflow.exceptions import AirflowException, AirflowProviderDeprecationWarning
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.common.deprecated import deprecated
from airflow.providers.google.marketing_platform.hooks.display_video import GoogleDisplayVideo360Hook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


@deprecated(
    planned_removal_date="September 01, 2025",
    use_instead="airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360CreateSDFDownloadTaskOperator",
    reason="Display & Video 360 API v2 has been deprecated and will be removed.",
    category=AirflowProviderDeprecationWarning,
)
class GoogleDisplayVideo360CreateQueryOperator(BaseOperator):
    """
    Creates a query.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        ``GoogleDisplayVideo360CreateQueryOperator``

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/bid-manager/v2/queries/create`

    :param body: Report object passed to the request's body as described here:
        https://developers.google.com/bid-manager/v2/queries#Query
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with the first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "body",
        "impersonation_chain",
    )
    template_ext: Sequence[str] = (".json",)

    def __init__(
        self,
        *,
        body: dict[str, Any],
        api_version: str = "v2",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.body = body
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def prepare_template(self) -> None:
        # If .json is passed then we have to read the file
        if isinstance(self.body, str) and self.body.endswith(".json"):
            with open(self.body) as file:
                self.body = json.load(file)

    def execute(self, context: Context) -> dict:
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info("Creating Display & Video 360 query.")
        response = hook.create_query(query=self.body)
        query_id = response["queryId"]
        context["task_instance"].xcom_push(key="query_id", value=query_id)
        self.log.info("Created query with ID: %s", query_id)
        return response


@deprecated(
    planned_removal_date="September 01, 2025",
    reason="Display & Video 360 API v2 has been deprecated and will be removed. "
    "Reports were replaced with SDF export task in v4 of API.",
    category=AirflowProviderDeprecationWarning,
)
class GoogleDisplayVideo360DeleteReportOperator(BaseOperator):
    """
    Deletes a stored query as well as the associated stored reports.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360DeleteReportOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/bid-manager/v2/queries/delete`

    :param report_id: Report ID to delete.
    :param report_name: Name of the report to delete.
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "report_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        report_id: str | None = None,
        report_name: str | None = None,
        api_version: str = "v2",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.report_id = report_id
        self.report_name = report_name
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

        if report_name and report_id:
            raise AirflowException("Use only one value - `report_name` or `report_id`.")

        if not (report_name or report_id):
            raise AirflowException("Provide one of the values: `report_name` or `report_id`.")

    def execute(self, context: Context) -> None:
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        if self.report_id:
            reports_ids_to_delete = [self.report_id]
        else:
            reports = hook.list_queries()
            reports_ids_to_delete = [
                report["queryId"] for report in reports if report["metadata"]["title"] == self.report_name
            ]

        for report_id in reports_ids_to_delete:
            self.log.info("Deleting report with id: %s", report_id)
            hook.delete_query(query_id=report_id)
            self.log.info("Report deleted.")


@deprecated(
    planned_removal_date="September 01, 2025",
    use_instead="airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360SDFtoGCSOperator",
    reason="Display & Video 360 API v2 has been deprecated and will be removed.",
    category=AirflowProviderDeprecationWarning,
)
class GoogleDisplayVideo360DownloadReportV2Operator(BaseOperator):
    """
    Retrieves a stored query.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360DownloadReportV2Operator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/bid-manager/v2/queries/get`

    :param report_id: Report ID to retrieve.
    :param bucket_name: The bucket to upload to.
    :param report_name: The report name to set when uploading the local file.
    :param chunk_size: File will be downloaded in chunks of this many bytes.
    :param gzip: Option to compress local file or file data for upload
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "query_id",
        "report_id",
        "bucket_name",
        "report_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        query_id: str,
        report_id: str,
        bucket_name: str,
        report_name: str | None = None,
        gzip: bool = True,
        chunk_size: int = 10 * 1024 * 1024,
        api_version: str = "v2",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.query_id = query_id
        self.report_id = report_id
        self.chunk_size = chunk_size
        self.gzip = gzip
        self.bucket_name = bucket_name
        self.report_name = report_name
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def _resolve_file_name(self, name: str) -> str:
        new_name = name if name.endswith(".csv") else f"{name}.csv"
        new_name = f"{new_name}.gz" if self.gzip else new_name
        return new_name

    @staticmethod
    def _set_bucket_name(name: str) -> str:
        bucket = name if not name.startswith("gs://") else name[5:]
        return bucket.strip("/")

    def execute(self, context: Context):
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        resource = hook.get_report(query_id=self.query_id, report_id=self.report_id)
        status = resource.get("metadata", {}).get("status", {}).get("state")
        if resource and status not in ["DONE", "FAILED"]:
            raise AirflowException(f"Report {self.report_id} for query {self.query_id} is still running")

        # If no custom report_name provided, use DV360 name
        file_url = resource["metadata"]["googleCloudStoragePath"]
        if urllib.parse.urlparse(file_url).scheme == "file":
            raise AirflowException("Accessing local file is not allowed in this operator")
        report_name = self.report_name or urlsplit(file_url).path.split("/")[-1]
        report_name = self._resolve_file_name(report_name)

        # Download the report
        self.log.info("Starting downloading report %s", self.report_id)
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            with urllib.request.urlopen(file_url) as response:  # nosec
                shutil.copyfileobj(response, temp_file, length=self.chunk_size)

            temp_file.flush()
            # Upload the local file to bucket
            bucket_name = self._set_bucket_name(self.bucket_name)
            gcs_hook.upload(
                bucket_name=bucket_name,
                object_name=report_name,
                gzip=self.gzip,
                filename=temp_file.name,
                mime_type="text/csv",
            )
        self.log.info(
            "Report %s was saved in bucket %s as %s.",
            self.report_id,
            self.bucket_name,
            report_name,
        )
        context["task_instance"].xcom_push(key="report_name", value=report_name)


@deprecated(
    planned_removal_date="September 01, 2025",
    use_instead="airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360CreateSDFDownloadTaskOperator",
    reason="Display & Video 360 API v2 has been deprecated and will be removed.",
    category=AirflowProviderDeprecationWarning,
)
class GoogleDisplayVideo360RunQueryOperator(BaseOperator):
    """
    Runs a stored query to generate a report.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360RunQueryOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/bid-manager/v2/queries/run`

    :param report_id: Report ID to run.
    :param parameters: Parameters for running a report as described here:
        https://developers.google.com/bid-manager/v2/queries/run
    :param api_version: The version of the api that will be requested for example 'v3'.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "query_id",
        "parameters",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        query_id: str,
        parameters: dict[str, Any] | None = None,
        api_version: str = "v2",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.query_id = query_id
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.parameters = parameters
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict:
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        self.log.info(
            "Running query %s with the following parameters:\n %s",
            self.query_id,
            self.parameters,
        )
        response = hook.run_query(query_id=self.query_id, params=self.parameters)
        context["task_instance"].xcom_push(key="query_id", value=response["key"]["queryId"])
        context["task_instance"].xcom_push(key="report_id", value=response["key"]["reportId"])
        return response


@deprecated(
    planned_removal_date="September 01, 2025",
    use_instead="airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360CreateSDFDownloadTaskOperator",
    reason="Display & Video 360 API v2 has been deprecated and will be removed.",
    category=AirflowProviderDeprecationWarning,
)
class GoogleDisplayVideo360DownloadLineItemsOperator(BaseOperator):
    """
    Retrieves line items in CSV format.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360DownloadLineItemsOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/bid-manager/v1.1/lineitems/downloadlineitems`

    :param request_body: dictionary with parameters that should be passed into.
            More information about it can be found here:
            https://developers.google.com/bid-manager/v1.1/lineitems/downloadlineitems
    """

    template_fields: Sequence[str] = (
        "request_body",
        "bucket_name",
        "object_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        request_body: dict[str, Any],
        bucket_name: str,
        object_name: str,
        gzip: bool = False,
        api_version: str = "v1.1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.request_body = request_body
        self.object_name = object_name
        self.bucket_name = bucket_name
        self.gzip = gzip
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> str:
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Retrieving report...")
        content: list[str] = hook.download_line_items(request_body=self.request_body)

        with tempfile.NamedTemporaryFile("w+") as temp_file:
            writer = csv.writer(temp_file)
            writer.writerows(content)
            temp_file.flush()
            gcs_hook.upload(
                bucket_name=self.bucket_name,
                object_name=self.object_name,
                filename=temp_file.name,
                mime_type="text/csv",
                gzip=self.gzip,
            )
        return f"{self.bucket_name}/{self.object_name}"


@deprecated(
    planned_removal_date="September 01, 2025",
    use_instead="airflow.providers.google.marketing_platform.operators.display_video.GoogleDisplayVideo360SDFtoGCSOperator",
    reason="Display & Video 360 API v2 has been deprecated and will be removed.",
    category=AirflowProviderDeprecationWarning,
)
class GoogleDisplayVideo360UploadLineItemsOperator(BaseOperator):
    """
    Uploads line items in CSV format.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360UploadLineItemsOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/bid-manager/v1.1/lineitems/uploadlineitems`

    :param request_body: request to upload line items.
    :param bucket_name: The bucket form data is downloaded.
    :param object_name: The object to fetch.
    :param filename: The filename to fetch.
    :param dry_run: Upload status without actually persisting the line items.
    """

    template_fields: Sequence[str] = (
        "bucket_name",
        "object_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        bucket_name: str,
        object_name: str,
        api_version: str = "v1.1",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Uploading file %s...")
        # Saving file in the temporary directory,
        # downloaded file from the GCS could be a 1GB size or even more
        with tempfile.NamedTemporaryFile("w+") as f:
            line_items = gcs_hook.download(
                bucket_name=self.bucket_name,
                object_name=self.object_name,
                filename=f.name,
            )
            f.flush()
            hook.upload_line_items(line_items=line_items)


class GoogleDisplayVideo360CreateSDFDownloadTaskOperator(BaseOperator):
    """
    Creates an SDF operation task.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360CreateSDFDownloadTaskOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/display-video/api/reference/rest`

    :param version: The SDF version of the downloaded file.
    :param partner_id: The ID of the partner to download SDF for.
    :param advertiser_id: The ID of the advertiser to download SDF for.
    :param parent_entity_filter: Filters on selected file types.
    :param id_filter: Filters on entities by their entity IDs.
    :param inventory_source_filter: Filters on Inventory Sources by their IDs.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "body_request",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        body_request: dict[str, Any],
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.body_request = body_request
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> dict[str, Any]:
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Creating operation for SDF download task...")
        operation = hook.create_sdf_download_operation(body_request=self.body_request)

        name = operation["name"]
        context["task_instance"].xcom_push(key="name", value=name)
        self.log.info("Created SDF operation with name: %s", name)

        return operation


class GoogleDisplayVideo360SDFtoGCSOperator(BaseOperator):
    """
    Download SDF media and save it in the Google Cloud Storage.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleDisplayVideo360SDFtoGCSOperator`

    .. seealso::
        Check also the official API docs:
        `https://developers.google.com/display-video/api/reference/rest`

    :param version: The SDF version of the downloaded file.
    :param partner_id: The ID of the partner to download SDF for.
    :param advertiser_id: The ID of the advertiser to download SDF for.
    :param parent_entity_filter: Filters on selected file types.
    :param id_filter: Filters on entities by their entity IDs.
    :param inventory_source_filter: Filters on Inventory Sources by their IDs.
    :param gcp_conn_id: The connection ID to use when fetching connection info.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "operation_name",
        "bucket_name",
        "object_name",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        operation_name: str,
        bucket_name: str,
        object_name: str,
        gzip: bool = False,
        api_version: str = "v4",
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.operation_name = operation_name
        self.bucket_name = bucket_name
        self.object_name = object_name
        self.gzip = gzip
        self.api_version = api_version
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> str:
        hook = GoogleDisplayVideo360Hook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        gcs_hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Retrieving operation...")
        operation_state = hook.get_sdf_download_operation(operation_name=self.operation_name)

        self.log.info("Creating file for upload...")
        media = hook.download_media(resource_name=operation_state["response"]["resourceName"])

        self.log.info("Sending file to the Google Cloud Storage...")
        with tempfile.TemporaryDirectory() as tmp_dir:
            zip_path = os.path.join(tmp_dir, "sdf.zip")

            # Download the ZIP
            with open(zip_path, "wb") as f:
                hook.download_content_from_request(f, media, chunk_size=1024 * 1024)

            # Extract CSV
            with zipfile.ZipFile(zip_path, "r") as zip_ref:
                zip_ref.extractall(tmp_dir)

            # Upload CSV file
            for fname in os.listdir(tmp_dir):
                if fname.endswith(".csv"):
                    gcs_hook.upload(
                        bucket_name=self.bucket_name,
                        object_name=self.object_name,
                        filename=os.path.join(tmp_dir, fname),
                        gzip=False,
                    )
        return f"{self.bucket_name}/{self.object_name}"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Base operator for SQL to GCS operators."""

from __future__ import annotations

import abc
import csv
import json
import os
from collections.abc import Sequence
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any

import pyarrow as pa
import pyarrow.parquet as pq

from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.openlineage.facet import OutputDataset
    from airflow.utils.context import Context


class BaseSQLToGCSOperator(BaseOperator):
    """
    Copy data from SQL to Google Cloud Storage in JSON, CSV, or Parquet format.

    :param sql: The SQL to execute.
    :param bucket: The bucket to upload to.
    :param filename: The filename to use as the object name when uploading
        to Google Cloud Storage. A ``{}`` should be specified in the filename
        to allow the operator to inject file numbers in cases where the
        file is split due to size.
    :param schema_filename: If set, the filename to use as the object name
        when uploading a .json file containing the BigQuery schema fields
        for the table that was dumped from the database.
    :param approx_max_file_size_bytes: This operator supports the ability
        to split large table dumps into multiple files (see notes in the
        filename param docs above). This param allows developers to specify the
        file size of the splits. Check https://cloud.google.com/storage/quotas
        to see the maximum allowed file size for a single object.
    :param export_format: Desired format of files to be exported. (json, csv or parquet)
    :param stringify_dict: Whether to dump Dictionary type objects
        (such as JSON columns) as a string. Applies only to CSV/JSON export format.
    :param field_delimiter: The delimiter to be used for CSV files.
    :param null_marker: The null marker to be used for CSV files.
    :param gzip: Option to compress file for upload (does not apply to schemas).
    :param schema: The schema to use, if any. Should be a list of dict or
        a str. Pass a string if using Jinja template, otherwise, pass a list of
        dict. Examples could be seen: https://cloud.google.com/bigquery/docs
        /schemas#specifying_a_json_schema_file
    :param gcp_conn_id: (Optional) The connection ID used to connect to Google Cloud.
    :param parameters: a parameters dict that is substituted at query runtime.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param upload_metadata: whether to upload the row count metadata as blob metadata
    :param exclude_columns: set of columns to exclude from transmission
    :param partition_columns: list of columns to use for file partitioning. In order to use
        this parameter, you must sort your dataset by partition_columns. Do this by
        passing an ORDER BY clause to the sql query. Files are uploaded to GCS as objects
        with a hive style partitioning directory structure (templated).
    :param write_on_empty: Optional parameter to specify whether to write a file if the
        export does not return any rows. Default is False so we will not write a file
        if the export returns no rows.
    :param parquet_row_group_size: The approximate number of rows in each row group
        when using parquet format. Using a large row group size can reduce the file size
        and improve the performance of reading the data, but it needs more memory to
        execute the operator. (default: 100000)
    """

    template_fields: Sequence[str] = (
        "sql",
        "bucket",
        "filename",
        "schema_filename",
        "schema",
        "parameters",
        "impersonation_chain",
        "partition_columns",
    )
    template_ext: Sequence[str] = (".sql",)
    template_fields_renderers = {"sql": "sql"}
    ui_color = "#a0e08c"

    def __init__(
        self,
        *,
        sql: str,
        bucket: str,
        filename: str,
        schema_filename: str | None = None,
        approx_max_file_size_bytes: int = 1900000000,
        export_format: str = "json",
        stringify_dict: bool = False,
        field_delimiter: str = ",",
        null_marker: str | None = None,
        gzip: bool = False,
        schema: str | list | None = None,
        parameters: dict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        upload_metadata: bool = False,
        exclude_columns: set | None = None,
        partition_columns: list | None = None,
        write_on_empty: bool = False,
        parquet_row_group_size: int = 100000,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if exclude_columns is None:
            exclude_columns = set()

        self.sql = sql
        self.bucket = bucket
        self.filename = filename
        self.schema_filename = schema_filename
        self.approx_max_file_size_bytes = approx_max_file_size_bytes
        self.export_format = export_format.lower()
        self.stringify_dict = stringify_dict
        self.field_delimiter = field_delimiter
        self.null_marker = null_marker
        self.gzip = gzip
        self.schema = schema
        self.parameters = parameters
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.upload_metadata = upload_metadata
        self.exclude_columns = exclude_columns
        self.partition_columns = partition_columns
        self.write_on_empty = write_on_empty
        self.parquet_row_group_size = parquet_row_group_size
        self._uploaded_file_names: list[str] = []

    def execute(self, context: Context):
        if self.partition_columns:
            self.log.info(
                "Found partition columns: %s. "
                "Assuming the SQL statement is properly sorted by these columns in "
                "ascending or descending order.",
                ",".join(self.partition_columns),
            )

        self.log.info("Executing query")
        cursor = self.query()

        # If a schema is set, create a BQ schema JSON file.
        if self.schema_filename:
            self.log.info("Writing local schema file")
            schema_file = self._write_local_schema_file(cursor)

            # Flush file before uploading
            schema_file["file_handle"].flush()

            self.log.info("Uploading schema file to GCS.")
            self._upload_to_gcs(schema_file)
            schema_file["file_handle"].close()

        counter = 0
        files = []
        total_row_count = 0
        total_files = 0
        self.log.info("Writing local data files")
        for file_to_upload in self._write_local_data_files(cursor):
            # Flush file before uploading
            file_to_upload["file_handle"].flush()

            self.log.info("Uploading chunk file #%d to GCS.", counter)
            self._upload_to_gcs(file_to_upload)

            self.log.info("Removing local file")
            file_to_upload["file_handle"].close()

            # Metadata to be outputted to Xcom
            total_row_count += file_to_upload["file_row_count"]
            total_files += 1
            files.append(
                {
                    "file_name": file_to_upload["file_name"],
                    "file_mime_type": file_to_upload["file_mime_type"],
                    "file_row_count": file_to_upload["file_row_count"],
                }
            )

            counter += 1

        file_meta = {
            "bucket": self.bucket,
            "total_row_count": total_row_count,
            "total_files": total_files,
            "files": files,
        }

        return file_meta

    def convert_types(self, schema, col_type_dict, row) -> list:
        """Convert values from DBAPI to output-friendly formats."""
        return [
            self.convert_type(value, col_type_dict.get(name), stringify_dict=self.stringify_dict)
            for name, value in zip(schema, row)
        ]

    @staticmethod
    def _write_rows_to_parquet(parquet_writer: pq.ParquetWriter, rows):
        rows_pydic: dict[str, list[Any]] = {col: [] for col in parquet_writer.schema.names}
        for row in rows:
            for cell, col in zip(row, parquet_writer.schema.names):
                rows_pydic[col].append(cell)
        tbl = pa.Table.from_pydict(rows_pydic, parquet_writer.schema)
        parquet_writer.write_table(tbl)

    def _write_local_data_files(self, cursor):
        """
        Take a cursor, and writes results to a local file.

        :return: A dictionary where keys are filenames to be used as object
            names in GCS, and values are file handles to local files that
            contain the data for the GCS objects.
        """
        org_schema = [schema_tuple[0] for schema_tuple in cursor.description]
        schema = [column for column in org_schema if column not in self.exclude_columns]

        col_type_dict = self._get_col_type_dict()
        file_no = 0
        file_mime_type = self._get_file_mime_type()
        file_to_upload, tmp_file_handle = self._get_file_to_upload(file_mime_type, file_no)

        if self.export_format == "csv":
            csv_writer = self._configure_csv_file(tmp_file_handle, schema)
        if self.export_format == "parquet":
            parquet_schema = self._convert_parquet_schema(cursor)
            parquet_writer = self._configure_parquet_file(tmp_file_handle, parquet_schema)
            rows_buffer = []

        prev_partition_values = None
        curr_partition_values = None
        for row in cursor:
            if self.partition_columns:
                row_dict = dict(zip(schema, row))
                curr_partition_values = tuple(
                    [row_dict.get(partition_column, "") for partition_column in self.partition_columns]
                )

                if prev_partition_values is None:
                    # We haven't set prev_partition_values before. Set to current
                    prev_partition_values = curr_partition_values

                elif prev_partition_values != curr_partition_values:
                    # If the partition values differ, write the current local file out
                    # Yield first before we write the current record
                    file_no += 1

                    if self.export_format == "parquet":
                        # Write out the remaining rows in the buffer
                        if rows_buffer:
                            self._write_rows_to_parquet(parquet_writer, rows_buffer)
                            rows_buffer = []
                        parquet_writer.close()

                    file_to_upload["partition_values"] = prev_partition_values
                    yield file_to_upload
                    file_to_upload, tmp_file_handle = self._get_file_to_upload(file_mime_type, file_no)
                    if self.export_format == "csv":
                        csv_writer = self._configure_csv_file(tmp_file_handle, schema)
                    if self.export_format == "parquet":
                        parquet_writer = self._configure_parquet_file(tmp_file_handle, parquet_schema)

                    # Reset previous to current after writing out the file
                    prev_partition_values = curr_partition_values

            # Incrementing file_row_count after partition yield ensures all rows are written
            file_to_upload["file_row_count"] += 1

            # Proceed to write the row to the localfile
            if self.export_format == "csv":
                row = self.convert_types(schema, col_type_dict, row)
                if self.null_marker is not None:
                    row = [value or self.null_marker for value in row]
                csv_writer.writerow(row)
            elif self.export_format == "parquet":
                row = self.convert_types(schema, col_type_dict, row)
                if self.null_marker is not None:
                    row = [value or self.null_marker for value in row]
                rows_buffer.append(row)
                if len(rows_buffer) >= self.parquet_row_group_size:
                    self._write_rows_to_parquet(parquet_writer, rows_buffer)
                    rows_buffer = []
            else:
                row = self.convert_types(schema, col_type_dict, row)
                row_dict = dict(zip(schema, row))

                json.dump(row_dict, tmp_file_handle, sort_keys=True, ensure_ascii=False)

                # Append newline to make dumps BigQuery compatible.
                tmp_file_handle.write("\n")

            # Stop if the file exceeds the file size limit.
            fppos = tmp_file_handle.tell()
            tmp_file_handle.seek(0, os.SEEK_END)
            file_size = tmp_file_handle.tell()
            tmp_file_handle.seek(fppos, os.SEEK_SET)

            if file_size >= self.approx_max_file_size_bytes:
                file_no += 1

                if self.export_format == "parquet":
                    # Write out the remaining rows in the buffer
                    if rows_buffer:
                        self._write_rows_to_parquet(parquet_writer, rows_buffer)
                        rows_buffer = []
                    parquet_writer.close()

                file_to_upload["partition_values"] = curr_partition_values
                yield file_to_upload
                file_to_upload, tmp_file_handle = self._get_file_to_upload(file_mime_type, file_no)
                if self.export_format == "csv":
                    csv_writer = self._configure_csv_file(tmp_file_handle, schema)
                if self.export_format == "parquet":
                    parquet_writer = self._configure_parquet_file(tmp_file_handle, parquet_schema)

        if self.export_format == "parquet":
            # Write out the remaining rows in the buffer
            if rows_buffer:
                self._write_rows_to_parquet(parquet_writer, rows_buffer)
                rows_buffer = []
            parquet_writer.close()
        # Last file may have 0 rows, don't yield if empty
        # However, if it is the first file and self.write_on_empty is True, then yield to write an empty file
        if file_to_upload["file_row_count"] > 0 or (file_no == 0 and self.write_on_empty):
            file_to_upload["partition_values"] = curr_partition_values
            yield file_to_upload

    def _get_file_to_upload(self, file_mime_type, file_no):
        """Return a dictionary that represents the file to upload."""
        tmp_file_handle = NamedTemporaryFile(mode="w", encoding="utf-8", delete=True)
        return (
            {
                "file_name": self.filename.format(file_no),
                "file_handle": tmp_file_handle,
                "file_mime_type": file_mime_type,
                "file_row_count": 0,
            },
            tmp_file_handle,
        )

    def _get_file_mime_type(self):
        if self.export_format == "csv":
            file_mime_type = "text/csv"
        elif self.export_format == "parquet":
            file_mime_type = "application/octet-stream"
        else:
            file_mime_type = "application/json"
        return file_mime_type

    def _configure_csv_file(self, file_handle, schema):
        """Configure a csv writer with the file_handle and write schema as headers for the new file."""
        csv_writer = csv.writer(file_handle, delimiter=self.field_delimiter)
        csv_writer.writerow(schema)
        return csv_writer

    def _configure_parquet_file(self, file_handle, parquet_schema) -> pq.ParquetWriter:
        parquet_writer = pq.ParquetWriter(file_handle.name, parquet_schema)
        return parquet_writer

    def _convert_parquet_schema(self, cursor):
        type_map = {
            "INTEGER": pa.int64(),
            "FLOAT": pa.float64(),
            "NUMERIC": pa.float64(),
            "BIGNUMERIC": pa.float64(),
            "BOOL": pa.bool_(),
            "STRING": pa.string(),
            "BYTES": pa.binary(),
            "DATE": pa.date32(),
            "DATETIME": pa.date64(),
            "TIMESTAMP": pa.timestamp("s"),
        }

        columns = [field[0] for field in cursor.description]
        bq_fields = [self.field_to_bigquery(field) for field in cursor.description]
        bq_types = [bq_field.get("type") if bq_field is not None else None for bq_field in bq_fields]
        pq_types = [type_map.get(bq_type, pa.string()) for bq_type in bq_types]
        parquet_schema = pa.schema(zip(columns, pq_types))
        return parquet_schema

    @abc.abstractmethod
    def query(self):
        """Execute DBAPI query."""

    @abc.abstractmethod
    def field_to_bigquery(self, field) -> dict[str, str]:
        """Convert a DBAPI field to BigQuery schema format."""

    @abc.abstractmethod
    def convert_type(self, value, schema_type, **kwargs):
        """Convert a value from DBAPI to output-friendly formats."""

    def _get_col_type_dict(self):
        """Return a dict of column name and column type based on self.schema if not None."""
        schema = []
        if isinstance(self.schema, str):
            schema = json.loads(self.schema)
        elif isinstance(self.schema, list):
            schema = self.schema
        elif self.schema is not None:
            self.log.warning("Using default schema due to unexpected type. Should be a string or list.")

        col_type_dict = {}
        try:
            col_type_dict = {col["name"]: col["type"] for col in schema}
        except KeyError:
            self.log.warning(
                "Using default schema due to missing name or type. Please "
                "refer to: https://cloud.google.com/bigquery/docs/schemas"
                "#specifying_a_json_schema_file"
            )
        return col_type_dict

    def _write_local_schema_file(self, cursor):
        """
        Take a cursor, and writes the BigQuery schema for the results to a local file system.

        Schema for database will be read from cursor if not specified.

        :return: A dictionary where key is a filename to be used as an object
            name in GCS, and values are file handles to local files that
            contains the BigQuery schema fields in .json format.
        """
        if self.schema:
            self.log.info("Using user schema")
            schema = self.schema
        else:
            self.log.info("Starts generating schema")
            schema = [
                self.field_to_bigquery(field)
                for field in cursor.description
                if field[0] not in self.exclude_columns
            ]

        if isinstance(schema, list):
            schema = json.dumps(schema, sort_keys=True)

        self.log.info("Using schema for %s", self.schema_filename)
        self.log.debug("Current schema: %s", schema)

        tmp_schema_file_handle = NamedTemporaryFile(mode="w", encoding="utf-8", delete=True)
        tmp_schema_file_handle.write(schema)
        schema_file_to_upload = {
            "file_name": self.schema_filename,
            "file_handle": tmp_schema_file_handle,
            "file_mime_type": "application/json",
        }
        return schema_file_to_upload

    def _upload_to_gcs(self, file_to_upload):
        """Upload a file (data split or schema .json file) to Google Cloud Storage."""
        hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        is_data_file = file_to_upload.get("file_name") != self.schema_filename
        metadata = None
        if is_data_file and self.upload_metadata:
            metadata = {"row_count": file_to_upload["file_row_count"]}

        object_name = file_to_upload.get("file_name")
        if is_data_file and self.partition_columns:
            # Add partition column values to object_name
            partition_values = file_to_upload.get("partition_values")
            head_path, tail_path = os.path.split(object_name)
            partition_subprefix = [
                f"{col}={val}" for col, val in zip(self.partition_columns, partition_values)
            ]
            object_name = os.path.join(head_path, *partition_subprefix, tail_path)

        hook.upload(
            self.bucket,
            object_name,
            file_to_upload.get("file_handle").name,
            mime_type=file_to_upload.get("file_mime_type"),
            gzip=self.gzip if is_data_file else False,
            metadata=metadata,
        )
        self._uploaded_file_names.append(object_name)

    def _get_openlineage_output_datasets(self) -> list[OutputDataset]:
        """Retrieve OpenLineage output datasets."""
        from airflow.providers.common.compat.openlineage.facet import OutputDataset
        from airflow.providers.google.cloud.openlineage.utils import extract_ds_name_from_gcs_path

        return [
            OutputDataset(
                namespace=f"gs://{self.bucket}",
                name=extract_ds_name_from_gcs_path(self.filename.split("{}", maxsplit=1)[0]),
            )
        ]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that displays interactions with Google Cloud Build.
"""

from __future__ import annotations

import os
from datetime import datetime
from typing import Any, cast

from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.exceptions import AirflowException
from airflow.models.dag import DAG
from airflow.models.xcom_arg import XComArg
from airflow.providers.google.cloud.operators.cloud_build import (
    CloudBuildCreateBuildTriggerOperator,
    CloudBuildDeleteBuildTriggerOperator,
    CloudBuildGetBuildTriggerOperator,
    CloudBuildListBuildTriggersOperator,
    CloudBuildRunBuildTriggerOperator,
    CloudBuildUpdateBuildTriggerOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "gcp_cloud_build_trigger"

# Repository with this name is expected created within the project $SYSTEM_TESTS_GCP_PROJECT
# If you'd like to run this system test locally, please
#   1. Create Cloud Source Repository
#   2. Push into a master branch the following file:
#   providers/google/tests/system/google/cloud/cloud_build/resources/example_cloud_build.yaml
GCP_SOURCE_REPOSITORY_NAME = "test-cloud-build-repository"

TRIGGER_NAME = f"cloud-build-trigger-{ENV_ID}".replace("_", "-")
PROJECT_NUMBER = "{{ task_instance.xcom_pull('get_project_number') }}"

# [START howto_operator_gcp_create_build_trigger_body]
create_build_trigger_body = {
    "name": TRIGGER_NAME,
    "trigger_template": {
        "project_id": PROJECT_ID,
        "repo_name": GCP_SOURCE_REPOSITORY_NAME,
        "branch_name": "master",
    },
    "filename": "example_cloud_build.yaml",
    "service_account": f"projects/{PROJECT_ID}/serviceAccounts/{PROJECT_NUMBER}-compute@developer.gserviceaccount.com",
}
# [END howto_operator_gcp_create_build_trigger_body]

# [START howto_operator_gcp_update_build_trigger_body]
update_build_trigger_body = {
    "name": TRIGGER_NAME,
    "trigger_template": {
        "project_id": PROJECT_ID,
        "repo_name": GCP_SOURCE_REPOSITORY_NAME,
        "branch_name": "master",
    },
    "filename": "example_cloud_build.yaml",
    "service_account": f"projects/{PROJECT_ID}/serviceAccounts/{PROJECT_NUMBER}-compute@developer.gserviceaccount.com",
}
# [END START howto_operator_gcp_update_build_trigger_body]

# [START howto_operator_create_build_from_repo_body]
create_build_from_repo_body: dict[str, Any] = {
    "source": {"repo_source": {"repo_name": GCP_SOURCE_REPOSITORY_NAME, "branch_name": "master"}},
    "steps": [{"name": "ubuntu", "args": ["echo", "Hello world", "sleep 200"]}],
}
# [END howto_operator_create_build_from_repo_body]


@task(task_id="get_project_number")
def get_project_number():
    """Helper function to retrieve the number of the project based on PROJECT_ID"""
    try:
        with build("cloudresourcemanager", "v1") as service:
            response = service.projects().get(projectId=PROJECT_ID).execute()
        return response["projectNumber"]
    except HttpError as exc:
        if exc.status_code == 403:
            raise AirflowException(
                "No project found with specified name, "
                "or caller does not have permissions to read specified project"
            )
        raise exc


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "cloud_build_trigger"],
) as dag:
    # [START howto_operator_create_build_trigger]
    create_build_trigger = CloudBuildCreateBuildTriggerOperator(
        task_id="create_build_trigger", project_id=PROJECT_ID, trigger=create_build_trigger_body
    )
    # [END howto_operator_create_build_trigger]

    build_trigger_id = cast("str", XComArg(create_build_trigger, key="id"))

    # [START howto_operator_run_build_trigger]
    run_build_trigger = CloudBuildRunBuildTriggerOperator(
        task_id="run_build_trigger",
        project_id=PROJECT_ID,
        trigger_id=build_trigger_id,
        source=create_build_from_repo_body["source"]["repo_source"],
    )
    # [END howto_operator_run_build_trigger]

    # [START howto_operator_update_build_trigger]
    update_build_trigger = CloudBuildUpdateBuildTriggerOperator(
        task_id="update_build_trigger",
        project_id=PROJECT_ID,
        trigger_id=build_trigger_id,
        trigger=update_build_trigger_body,
    )
    # [END howto_operator_update_build_trigger]

    # [START howto_operator_get_build_trigger]
    get_build_trigger = CloudBuildGetBuildTriggerOperator(
        task_id="get_build_trigger",
        project_id=PROJECT_ID,
        trigger_id=build_trigger_id,
    )
    # [END howto_operator_get_build_trigger]

    # [START howto_operator_delete_build_trigger]
    delete_build_trigger = CloudBuildDeleteBuildTriggerOperator(
        task_id="delete_build_trigger",
        project_id=PROJECT_ID,
        trigger_id=build_trigger_id,
    )
    # [END howto_operator_delete_build_trigger]
    delete_build_trigger.trigger_rule = TriggerRule.ALL_DONE

    # [START howto_operator_list_build_triggers]
    list_build_triggers = CloudBuildListBuildTriggersOperator(
        task_id="list_build_triggers",
        project_id=PROJECT_ID,
        location="global",
        page_size=5,
    )
    # [END howto_operator_list_build_triggers]

    (
        # TEST SETUP
        get_project_number()
        # TEST BODY
        >> create_build_trigger
        >> run_build_trigger
        >> update_build_trigger
        >> get_build_trigger
        # TEST TEARDOWN
        >> delete_build_trigger
        >> list_build_triggers
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import sys
from pathlib import Path

source = sys.argv[1]
destination = sys.argv[2]
timespan_start = sys.argv[3]
timespan_end = sys.argv[4]

print(sys.argv)
print(f"Running script, called with source: {source}, destination: {destination}")
print(f"timespan_start: {timespan_start}, timespan_end: {timespan_end}")

path = next(path for path in Path(source).rglob("*") if path.is_file())
content = path.read_text().upper()
print(content)
Path(destination, "output.txt").write_text(content)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from unittest import mock

from airflow.providers.google.suite.sensors.drive import GoogleDriveFileExistenceSensor

TEST_FOLDER_ID = os.environ.get("GCP_GDRIVE_FOLDER_ID", "abcd1234")
TEST_FILE_NAME = os.environ.get("GCP_GDRIVE_DRIVE_ID", "gdrive_to_gcs_file.txt")
TEST_DRIVE_ID = os.environ.get("GCP_GDRIVE_TO_GCS_FILE_NAME", "abcd1234")
TEST_GCP_CONN_ID = "google_cloud_default"
TEST_IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]


class TestGoogleDriveFileSensor:
    @mock.patch("airflow.providers.google.suite.sensors.drive.GoogleDriveHook")
    def test_should_pass_argument_to_hook(self, mock_hook):
        task = GoogleDriveFileExistenceSensor(
            task_id="task-id",
            folder_id=TEST_FOLDER_ID,
            file_name=TEST_FILE_NAME,
            drive_id=TEST_DRIVE_ID,
            gcp_conn_id=TEST_GCP_CONN_ID,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.exists.return_value = True

        result = task.poke(mock.MagicMock())
        assert result

        mock_hook.assert_called_once_with(
            gcp_conn_id=TEST_GCP_CONN_ID,
            impersonation_chain=TEST_IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.exists.assert_called_once_with(
            folder_id=TEST_FOLDER_ID, file_name=TEST_FILE_NAME, drive_id=TEST_DRIVE_ID
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to use Dataplex Catalog.
"""

from __future__ import annotations

import datetime
import os

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.dataplex import (
    DataplexCatalogCreateAspectTypeOperator,
    DataplexCatalogCreateEntryGroupOperator,
    DataplexCatalogCreateEntryOperator,
    DataplexCatalogCreateEntryTypeOperator,
    DataplexCatalogDeleteAspectTypeOperator,
    DataplexCatalogDeleteEntryGroupOperator,
    DataplexCatalogDeleteEntryOperator,
    DataplexCatalogDeleteEntryTypeOperator,
    DataplexCatalogGetAspectTypeOperator,
    DataplexCatalogGetEntryGroupOperator,
    DataplexCatalogGetEntryOperator,
    DataplexCatalogGetEntryTypeOperator,
    DataplexCatalogListAspectTypesOperator,
    DataplexCatalogListEntriesOperator,
    DataplexCatalogListEntryGroupsOperator,
    DataplexCatalogListEntryTypesOperator,
    DataplexCatalogLookupEntryOperator,
    DataplexCatalogSearchEntriesOperator,
    DataplexCatalogUpdateAspectTypeOperator,
    DataplexCatalogUpdateEntryGroupOperator,
    DataplexCatalogUpdateEntryOperator,
    DataplexCatalogUpdateEntryTypeOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "dataplex_catalog"
GCP_LOCATION = "us-central1"

ENTRY_GROUP_NAME = f"{DAG_ID}_entry_group_{ENV_ID}".replace("_", "-")
# [START howto_dataplex_entry_group_configuration]
ENTRY_GROUP_BODY = {"display_name": "Display Name", "description": "Some description"}
# [END howto_dataplex_entry_group_configuration]

ENTRY_TYPE_NAME = f"{DAG_ID}_entry_type_{ENV_ID}".replace("_", "-")
# [START howto_dataplex_entry_type_configuration]
ENTRY_TYPE_BODY = {"display_name": "Display Name", "description": "Some description"}
# [END howto_dataplex_entry_type_configuration]

ASPECT_TYPE_NAME = f"{DAG_ID}_aspect_type_{ENV_ID}".replace("_", "-")
# [START howto_dataplex_aspect_type_configuration]
ASPECT_TYPE_BODY = {
    "display_name": "Sample AspectType",
    "description": "A simple AspectType for demonstration purposes.",
    "metadata_template": {
        "name": "sample_field",
        "type": "record",
        "annotations": {
            "display_name": "Sample Field",
            "description": "A sample field within the AspectType.",
        },
    },
}
# [END howto_dataplex_aspect_type_configuration]

ENTRY_NAME = f"{DAG_ID}_entry_{ENV_ID}".replace("_", "-")
# [START howto_dataplex_entry_configuration]
ENTRY_BODY = {
    "name": f"projects/{PROJECT_ID}/locations/{GCP_LOCATION}/entryGroups/{ENTRY_GROUP_NAME}/entries/{ENTRY_NAME}",
    "entry_type": f"projects/{PROJECT_ID}/locations/{GCP_LOCATION}/entryTypes/{ENTRY_TYPE_NAME}",
}
# [END howto_dataplex_entry_configuration]

with DAG(
    DAG_ID,
    start_date=datetime.datetime(2021, 1, 1),
    schedule="@once",
    tags=["example", "dataplex_catalog"],
) as dag:
    # [START howto_operator_dataplex_catalog_create_entry_group]
    create_entry_group = DataplexCatalogCreateEntryGroupOperator(
        task_id="create_entry_group",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_group_id=ENTRY_GROUP_NAME,
        entry_group_configuration=ENTRY_GROUP_BODY,
        validate_request=False,
    )
    # [END howto_operator_dataplex_catalog_create_entry_group]

    # [START howto_operator_dataplex_catalog_create_entry_type]
    create_entry_type = DataplexCatalogCreateEntryTypeOperator(
        task_id="create_entry_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_type_id=ENTRY_TYPE_NAME,
        entry_type_configuration=ENTRY_TYPE_BODY,
        validate_request=False,
    )
    # [END howto_operator_dataplex_catalog_create_entry_type]

    # [START howto_operator_dataplex_catalog_create_aspect_type]
    create_aspect_type = DataplexCatalogCreateAspectTypeOperator(
        task_id="create_aspect_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        aspect_type_id=ASPECT_TYPE_NAME,
        aspect_type_configuration=ASPECT_TYPE_BODY,
        validate_request=False,
    )
    # [END howto_operator_dataplex_catalog_create_aspect_type]

    # [START howto_operator_dataplex_catalog_create_entry]
    create_entry = DataplexCatalogCreateEntryOperator(
        task_id="create_entry",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_id=ENTRY_NAME,
        entry_group_id=ENTRY_GROUP_NAME,
        entry_configuration=ENTRY_BODY,
    )
    # [END howto_operator_dataplex_catalog_create_entry]

    # [START howto_operator_dataplex_catalog_get_entry_group]
    get_entry_group = DataplexCatalogGetEntryGroupOperator(
        task_id="get_entry_group",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_group_id=ENTRY_GROUP_NAME,
    )
    # [END howto_operator_dataplex_catalog_get_entry_group]

    # [START howto_operator_dataplex_catalog_get_entry_type]
    get_entry_type = DataplexCatalogGetEntryTypeOperator(
        task_id="get_entry_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_type_id=ENTRY_TYPE_NAME,
    )
    # [END howto_operator_dataplex_catalog_get_entry_type]

    # [START howto_operator_dataplex_catalog_get_aspect_type]
    get_aspect_type = DataplexCatalogGetAspectTypeOperator(
        task_id="get_aspect_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        aspect_type_id=ASPECT_TYPE_NAME,
    )
    # [END howto_operator_dataplex_catalog_get_aspect_type]

    # [START howto_operator_dataplex_catalog_get_entry]
    get_entry = DataplexCatalogGetEntryOperator(
        task_id="get_entry",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_id=ENTRY_NAME,
        entry_group_id=ENTRY_GROUP_NAME,
    )
    # [END howto_operator_dataplex_catalog_get_entry]

    # [START howto_operator_dataplex_catalog_list_entry_groups]
    list_entry_group = DataplexCatalogListEntryGroupsOperator(
        task_id="list_entry_group",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        order_by="name",
        filter_by='display_name = "Display Name"',
    )
    # [END howto_operator_dataplex_catalog_list_entry_groups]

    # [START howto_operator_dataplex_catalog_list_entry_types]
    list_entry_type = DataplexCatalogListEntryTypesOperator(
        task_id="list_entry_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        order_by="name",
        filter_by='display_name = "Display Name"',
    )
    # [END howto_operator_dataplex_catalog_list_entry_types]

    # [START howto_operator_dataplex_catalog_list_aspect_types]
    list_aspect_type = DataplexCatalogListAspectTypesOperator(
        task_id="list_aspect_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        order_by="name",
        filter_by='display_name = "Display Name"',
    )
    # [END howto_operator_dataplex_catalog_list_aspect_types]

    # [START howto_operator_dataplex_catalog_list_entries]
    list_entry = DataplexCatalogListEntriesOperator(
        task_id="list_entry",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_group_id=ENTRY_GROUP_NAME,
    )
    # [END howto_operator_dataplex_catalog_list_entries]

    # [START howto_operator_dataplex_catalog_update_entry_group]
    update_entry_group = DataplexCatalogUpdateEntryGroupOperator(
        task_id="update_entry_group",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_group_id=ENTRY_GROUP_NAME,
        entry_group_configuration={"display_name": "Updated Display Name"},
        update_mask=["display_name"],
    )
    # [END howto_operator_dataplex_catalog_update_entry_group]

    # [START howto_operator_dataplex_catalog_update_entry_type]
    update_entry_type = DataplexCatalogUpdateEntryTypeOperator(
        task_id="update_entry_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_type_id=ENTRY_TYPE_NAME,
        entry_type_configuration={"display_name": "Updated Display Name"},
        update_mask=["display_name"],
    )
    # [END howto_operator_dataplex_catalog_update_entry_type]

    # [START howto_operator_dataplex_catalog_update_aspect_type]
    update_aspect_type = DataplexCatalogUpdateAspectTypeOperator(
        task_id="update_aspect_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        aspect_type_id=ASPECT_TYPE_NAME,
        aspect_type_configuration={"display_name": "Updated Display Name"},
        update_mask=["display_name"],
    )
    # [END howto_operator_dataplex_catalog_update_aspect_type]

    # [START howto_operator_dataplex_catalog_update_entry]
    update_entry = DataplexCatalogUpdateEntryOperator(
        task_id="update_entry",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_id=ENTRY_NAME,
        entry_group_id=ENTRY_GROUP_NAME,
        entry_configuration={
            "fully_qualified_name": f"dataplex:{PROJECT_ID}.{GCP_LOCATION}.{ENTRY_GROUP_NAME}.some-entry"
        },
        update_mask=["fully_qualified_name"],
    )
    # [END howto_operator_dataplex_catalog_update_entry]

    # [START howto_operator_dataplex_catalog_search_entry]
    search_entry = DataplexCatalogSearchEntriesOperator(
        task_id="search_entry",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        query=f"name={ENTRY_NAME}",
    )
    # [END howto_operator_dataplex_catalog_search_entry]

    # [START howto_operator_dataplex_catalog_lookup_entry]
    lookup_entry = DataplexCatalogLookupEntryOperator(
        task_id="lookup_entry",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_id=ENTRY_NAME,
        entry_group_id=ENTRY_GROUP_NAME,
    )
    # [END howto_operator_dataplex_catalog_lookup_entry]

    # [START howto_operator_dataplex_catalog_delete_entry_group]
    delete_entry_group = DataplexCatalogDeleteEntryGroupOperator(
        task_id="delete_entry_group",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_group_id=ENTRY_GROUP_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_dataplex_catalog_delete_entry_group]

    # [START howto_operator_dataplex_catalog_delete_entry_type]
    delete_entry_type = DataplexCatalogDeleteEntryTypeOperator(
        task_id="delete_entry_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_type_id=ENTRY_TYPE_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_dataplex_catalog_delete_entry_type]

    # [START howto_operator_dataplex_catalog_delete_aspect_type]
    delete_aspect_type = DataplexCatalogDeleteAspectTypeOperator(
        task_id="delete_aspect_type",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        aspect_type_id=ASPECT_TYPE_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_dataplex_catalog_delete_aspect_type]

    # [START howto_operator_dataplex_catalog_delete_entry]
    delete_entry = DataplexCatalogDeleteEntryOperator(
        task_id="delete_entry",
        project_id=PROJECT_ID,
        location=GCP_LOCATION,
        entry_id=ENTRY_NAME,
        entry_group_id=ENTRY_GROUP_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )
    # [END howto_operator_dataplex_catalog_delete_entry]

    (
        create_entry_group
        >> create_entry_type
        >> create_aspect_type
        >> create_entry
        >> search_entry
        >> [get_entry_group, get_entry_type, get_aspect_type, get_entry]
        >> lookup_entry
        >> [list_entry, list_entry_group, list_entry_type, list_aspect_type]
        >> update_entry_group
        >> update_entry_type
        >> update_entry
        >> update_aspect_type
        >> delete_entry
        >> delete_aspect_type
        >> delete_entry_type
        >> delete_entry_group
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import csv
from collections.abc import Sequence
from operator import attrgetter
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING

from airflow.providers.google.ads.hooks.ads import GoogleAdsHook
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.utils.context import Context


class GoogleAdsToGcsOperator(BaseOperator):
    """
    Fetch daily results from the Google Ads API for 1-n clients.

    Converts and saves the data as a temporary CSV file Uploads the CSV to
    Google Cloud Storage.

    .. seealso::
        For more information on the Google Ads API, take a look at the API docs:
        https://developers.google.com/google-ads/api/docs/start

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GoogleAdsToGcsOperator`

    :param client_ids: Google Ads client IDs to query
    :param query: Google Ads Query Language API query
    :param attributes: List of Google Ads Row attributes to extract
    :param bucket: The GCS bucket to upload to
    :param obj: GCS path to save the object. Must be the full file path (ex. `path/to/file.txt`)
    :param gcp_conn_id: Airflow Google Cloud connection ID
    :param google_ads_conn_id: Airflow Google Ads connection ID
    :param gzip: Option to compress local file or file data for upload
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param api_version: Optional Google Ads API version to use.
    """

    template_fields: Sequence[str] = (
        "client_ids",
        "query",
        "attributes",
        "bucket",
        "obj",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        client_ids: list[str],
        query: str,
        attributes: list[str],
        bucket: str,
        obj: str,
        gcp_conn_id: str = "google_cloud_default",
        google_ads_conn_id: str = "google_ads_default",
        gzip: bool = False,
        impersonation_chain: str | Sequence[str] | None = None,
        api_version: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.client_ids = client_ids
        self.query = query
        self.attributes = attributes
        self.bucket = bucket
        self.obj = obj
        self.gcp_conn_id = gcp_conn_id
        self.google_ads_conn_id = google_ads_conn_id
        self.gzip = gzip
        self.impersonation_chain = impersonation_chain
        self.api_version = api_version

    def execute(self, context: Context) -> None:
        service = GoogleAdsHook(
            gcp_conn_id=self.gcp_conn_id,
            google_ads_conn_id=self.google_ads_conn_id,
            api_version=self.api_version,
        )

        rows = service.search(client_ids=self.client_ids, query=self.query)

        try:
            getter = attrgetter(*self.attributes)
            converted_rows = [getter(row) for row in rows]
        except Exception as e:
            self.log.error("An error occurred in converting the Google Ad Rows. \n Error %s", e)
            raise

        with NamedTemporaryFile("w", suffix=".csv") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerows(converted_rows)
            csvfile.flush()

            hook = GCSHook(gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain)
            hook.upload(
                bucket_name=self.bucket,
                object_name=self.obj,
                filename=csvfile.name,
                gzip=self.gzip,
            )
            self.log.info("%s uploaded to GCS", self.obj)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Example Airflow DAG for testing Google Dataflow Beam Pipeline Operator with Asynchronous Python in the deferrable mode."""

from __future__ import annotations

import os
from collections.abc import Callable
from datetime import datetime

from airflow.exceptions import AirflowException
from airflow.models.dag import DAG
from airflow.providers.apache.beam.hooks.beam import BeamRunnerType
from airflow.providers.apache.beam.operators.beam import BeamRunPythonPipelineOperator
from airflow.providers.google.cloud.hooks.dataflow import DataflowJobStatus
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.sensors.dataflow import (
    DataflowJobAutoScalingEventsSensor,
    DataflowJobMessagesSensor,
    DataflowJobMetricsSensor,
    DataflowJobStatusSensor,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "dataflow_sensors_deferrable"

RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}"

GCS_TMP = f"gs://{BUCKET_NAME}/temp/"
GCS_STAGING = f"gs://{BUCKET_NAME}/staging/"
GCS_OUTPUT = f"gs://{BUCKET_NAME}/output"
GCS_PYTHON_SCRIPT = f"gs://{RESOURCE_DATA_BUCKET}/dataflow/python/wordcount_debugging.py"
LOCATION = "europe-west3"

default_args = {
    "dataflow_default_options": {
        "tempLocation": GCS_TMP,
        "stagingLocation": GCS_STAGING,
    }
}


with DAG(
    DAG_ID,
    default_args=default_args,
    schedule="@once",
    start_date=datetime(2024, 1, 1),
    catchup=False,
    tags=["example", "dataflow"],
) as dag:
    create_bucket = GCSCreateBucketOperator(task_id="create_bucket", bucket_name=BUCKET_NAME)

    start_beam_python_pipeline = BeamRunPythonPipelineOperator(
        task_id="start_beam_python_pipeline",
        runner=BeamRunnerType.DataflowRunner,
        py_file=GCS_PYTHON_SCRIPT,
        py_options=[],
        pipeline_options={
            "output": GCS_OUTPUT,
        },
        py_requirements=["apache-beam[gcp]==2.59.0"],
        py_interpreter="python3",
        py_system_site_packages=False,
        dataflow_config={
            "job_name": "start_beam_python_pipeline",
            "location": LOCATION,
            "wait_until_finished": False,
        },
    )

    # [START howto_sensor_wait_for_job_status_deferrable]
    wait_for_beam_python_pipeline_job_status_def = DataflowJobStatusSensor(
        task_id="wait_for_beam_python_pipeline_job_status_def",
        job_id="{{task_instance.xcom_pull('start_beam_python_pipeline')['dataflow_job_id']}}",
        expected_statuses=DataflowJobStatus.JOB_STATE_DONE,
        location=LOCATION,
        deferrable=True,
    )
    # [END howto_sensor_wait_for_job_status_deferrable]

    # [START howto_sensor_wait_for_job_metric_deferrable]
    def check_metric_scalar_gte(metric_name: str, value: int) -> Callable:
        """Check is metric greater than equals to given value."""

        def callback(metrics: list[dict]) -> bool:
            dag.log.info("Looking for '%s' >= %d", metric_name, value)
            for metric in metrics:
                context = metric.get("name", {}).get("context", {})
                original_name = context.get("original_name", "")
                tentative = context.get("tentative", "")
                if original_name == "Service-cpu_num_seconds" and not tentative:
                    return metric["scalar"] >= value
            raise AirflowException(f"Metric '{metric_name}' not found in metrics")

        return callback

    wait_for_beam_python_pipeline_job_metric_def = DataflowJobMetricsSensor(
        task_id="wait_for_beam_python_pipeline_job_metric_def",
        job_id="{{task_instance.xcom_pull('start_beam_python_pipeline')['dataflow_job_id']}}",
        location=LOCATION,
        callback=check_metric_scalar_gte(metric_name="Service-cpu_num_seconds", value=100),
        fail_on_terminal_state=False,
        deferrable=True,
    )
    # [END howto_sensor_wait_for_job_metric_deferrable]

    # [START howto_sensor_wait_for_job_message_deferrable]
    def check_job_message(messages: list[dict]) -> bool:
        """Check job message."""
        for message in messages:
            if "Adding workflow start and stop steps." in message.get("messageText", ""):
                return True
        return False

    wait_for_beam_python_pipeline_job_message_def = DataflowJobMessagesSensor(
        task_id="wait_for_beam_python_pipeline_job_message_def",
        job_id="{{task_instance.xcom_pull('start_beam_python_pipeline')['dataflow_job_id']}}",
        location=LOCATION,
        callback=check_job_message,
        fail_on_terminal_state=False,
        deferrable=True,
    )
    # [END howto_sensor_wait_for_job_message_deferrable]

    # [START howto_sensor_wait_for_job_autoscaling_event_deferrable]
    def check_autoscaling_event(autoscaling_events: list[dict]) -> bool:
        """Check autoscaling event."""
        for autoscaling_event in autoscaling_events:
            if "Worker pool started." in autoscaling_event.get("description", {}).get("messageText", ""):
                return True
        return False

    wait_for_beam_python_pipeline_job_autoscaling_event_def = DataflowJobAutoScalingEventsSensor(
        task_id="wait_for_beam_python_pipeline_job_autoscaling_event_def",
        job_id="{{task_instance.xcom_pull('start_beam_python_pipeline')['dataflow_job_id']}}",
        location=LOCATION,
        callback=check_autoscaling_event,
        fail_on_terminal_state=False,
        deferrable=True,
    )
    # [END howto_sensor_wait_for_job_autoscaling_event_deferrable]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket", bucket_name=BUCKET_NAME, trigger_rule=TriggerRule.ALL_DONE
    )

    (
        # TEST SETUP
        create_bucket
        >> start_beam_python_pipeline
        # TEST BODY
        >> [
            wait_for_beam_python_pipeline_job_status_def,
            wait_for_beam_python_pipeline_job_metric_def,
            wait_for_beam_python_pipeline_job_message_def,
            wait_for_beam_python_pipeline_job_autoscaling_event_def,
        ]
        # TEST TEARDOWN
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "teardown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.operators.datastore import (
    CloudDatastoreAllocateIdsOperator,
    CloudDatastoreBeginTransactionOperator,
    CloudDatastoreCommitOperator,
    CloudDatastoreDeleteOperationOperator,
    CloudDatastoreExportEntitiesOperator,
    CloudDatastoreGetOperationOperator,
    CloudDatastoreImportEntitiesOperator,
    CloudDatastoreRollbackOperator,
    CloudDatastoreRunQueryOperator,
)

HOOK_PATH = "airflow.providers.google.cloud.operators.datastore.DatastoreHook"
PROJECT_ID = "test-project"
CONN_ID = "test-gcp-conn-id"
BODY = {"key", "value"}
TRANSACTION = "transaction-name"
BUCKET = "gs://test-bucket"
OUTPUT_URL = f"{BUCKET}/entities_export_name/entities_export_name.overall_export_metadata"
FILE = "filename"
OPERATION_ID = "1234"


class TestCloudDatastoreExportEntitiesOperator:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        mock_hook.return_value.export_to_storage_bucket.return_value = {"name": OPERATION_ID}
        mock_hook.return_value.poll_operation_until_done.return_value = {
            "metadata": {"common": {"state": "SUCCESSFUL"}},
            "response": {
                "outputUrl": OUTPUT_URL,
            },
        }

        op = CloudDatastoreExportEntitiesOperator(
            task_id="test_task",
            datastore_conn_id=CONN_ID,
            project_id=PROJECT_ID,
            bucket=BUCKET,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.export_to_storage_bucket.assert_called_once_with(
            project_id=PROJECT_ID,
            bucket=BUCKET,
            entity_filter=None,
            labels=None,
            namespace=None,
        )

        mock_hook.return_value.poll_operation_until_done.assert_called_once_with(OPERATION_ID, 10)


class TestCloudDatastoreImportEntitiesOperator:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        mock_hook.return_value.import_from_storage_bucket.return_value = {"name": OPERATION_ID}
        mock_hook.return_value.poll_operation_until_done.return_value = {
            "metadata": {"common": {"state": "SUCCESSFUL"}}
        }

        op = CloudDatastoreImportEntitiesOperator(
            task_id="test_task",
            datastore_conn_id=CONN_ID,
            project_id=PROJECT_ID,
            bucket=BUCKET,
            file=FILE,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})

        mock_hook.assert_called_once_with(CONN_ID, impersonation_chain=None)
        mock_hook.return_value.import_from_storage_bucket.assert_called_once_with(
            project_id=PROJECT_ID,
            bucket=BUCKET,
            file=FILE,
            entity_filter=None,
            labels=None,
            namespace=None,
        )

        mock_hook.return_value.export_to_storage_bucketassert_called_once_with(OPERATION_ID, 10)


class TestCloudDatastoreAllocateIds:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        partial_keys = [1, 2, 3]
        op = CloudDatastoreAllocateIdsOperator(
            task_id="test_task",
            gcp_conn_id=CONN_ID,
            project_id=PROJECT_ID,
            partial_keys=partial_keys,
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.allocate_ids.assert_called_once_with(
            project_id=PROJECT_ID, partial_keys=partial_keys
        )


class TestCloudDatastoreBeginTransaction:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        op = CloudDatastoreBeginTransactionOperator(
            task_id="test_task",
            gcp_conn_id=CONN_ID,
            project_id=PROJECT_ID,
            transaction_options=BODY,
        )
        op.execute({})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.begin_transaction.assert_called_once_with(
            project_id=PROJECT_ID, transaction_options=BODY
        )


class TestCloudDatastoreCommit:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        op = CloudDatastoreCommitOperator(
            task_id="test_task", gcp_conn_id=CONN_ID, project_id=PROJECT_ID, body=BODY
        )
        op.execute(context={"ti": mock.MagicMock(), "task": mock.MagicMock()})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.commit.assert_called_once_with(project_id=PROJECT_ID, body=BODY)


class TestCloudDatastoreDeleteOperation:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        op = CloudDatastoreDeleteOperationOperator(task_id="test_task", gcp_conn_id=CONN_ID, name=TRANSACTION)
        op.execute({})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.delete_operation.assert_called_once_with(name=TRANSACTION)


class TestCloudDatastoreGetOperation:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        op = CloudDatastoreGetOperationOperator(task_id="test_task", gcp_conn_id=CONN_ID, name=TRANSACTION)
        op.execute({})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.get_operation.assert_called_once_with(name=TRANSACTION)


class TestCloudDatastoreRollback:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        op = CloudDatastoreRollbackOperator(
            task_id="test_task",
            gcp_conn_id=CONN_ID,
            project_id=PROJECT_ID,
            transaction=TRANSACTION,
        )
        op.execute({})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.rollback.assert_called_once_with(
            project_id=PROJECT_ID, transaction=TRANSACTION
        )


class TestCloudDatastoreRunQuery:
    @mock.patch(HOOK_PATH)
    def test_execute(self, mock_hook):
        op = CloudDatastoreRunQueryOperator(
            task_id="test_task", gcp_conn_id=CONN_ID, project_id=PROJECT_ID, body=BODY
        )
        op.execute({})

        mock_hook.assert_called_once_with(gcp_conn_id=CONN_ID, impersonation_chain=None)
        mock_hook.return_value.run_query.assert_called_once_with(project_id=PROJECT_ID, body=BODY)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from airflow.providers.google.cloud.transfers.facebook_ads_to_gcs import FacebookAdsReportToGcsOperator

GCS_BUCKET = "airflow_bucket_fb"
GCS_OBJ_PATH = "Temp/this_is_my_report_json.json"
GCS_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
FACEBOOK_ADS_CONN_ID = "facebook_default"
FIELDS = [
    "campaign_name",
    "campaign_id",
    "ad_id",
    "clicks",
    "impressions",
]
PARAMETERS = {"level": "ad", "date_preset": "yesterday"}
ACCOUNT_ID_1 = "act_12345"
ACCOUNT_ID_2 = "act_12346"
GCS_OBJ_PATH_1 = "Temp/act_12345_this_is_my_report_json.json"
GCS_OBJ_PATH_2 = "Temp/act_12346_this_is_my_report_json.json"
FACEBOOK_RETURN_VALUE = {
    ACCOUNT_ID_1: [
        {
            "campaign_name": "abcd",
            "campaign_id": "abcd",
            "ad_id": "abcd",
            "clicks": "2",
            "impressions": "2",
        }
    ],
    ACCOUNT_ID_2: [
        {
            "campaign_name": "abcd",
            "campaign_id": "abcd",
            "ad_id": "abcd",
            "clicks": "2",
            "impressions": "2",
        }
    ],
}


class TestFacebookAdsReportToGcsOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.facebook_ads_to_gcs.FacebookAdsReportingHook")
    @mock.patch("airflow.providers.google.cloud.transfers.facebook_ads_to_gcs.GCSHook")
    def test_execute(self, mock_gcs_hook, mock_ads_hook):
        mock_ads_hook.return_value.bulk_facebook_report.return_value = FACEBOOK_RETURN_VALUE
        op = FacebookAdsReportToGcsOperator(
            facebook_conn_id=FACEBOOK_ADS_CONN_ID,
            fields=FIELDS,
            parameters=PARAMETERS,
            object_name=GCS_OBJ_PATH,
            bucket_name=GCS_BUCKET,
            task_id="run_operator",
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute({})
        mock_ads_hook.assert_called_once_with(facebook_conn_id=FACEBOOK_ADS_CONN_ID, api_version=None)
        mock_ads_hook.return_value.bulk_facebook_report.assert_called_once_with(
            params=PARAMETERS, fields=FIELDS
        )
        mock_gcs_hook.assert_called_once_with(
            gcp_conn_id=GCS_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_gcs_hook.return_value.upload.assert_called_once_with(
            bucket_name=GCS_BUCKET, object_name=GCS_OBJ_PATH, filename=mock.ANY, gzip=False
        )

    @mock.patch("airflow.providers.google.cloud.transfers.facebook_ads_to_gcs.FacebookAdsReportingHook")
    @mock.patch("airflow.providers.google.cloud.transfers.facebook_ads_to_gcs.GCSHook")
    def test_execute_with_upload(self, mock_gcs_hook, mock_ads_hook):
        mock_ads_hook.return_value.bulk_facebook_report.return_value = FACEBOOK_RETURN_VALUE
        op = FacebookAdsReportToGcsOperator(
            facebook_conn_id=FACEBOOK_ADS_CONN_ID,
            fields=FIELDS,
            parameters=PARAMETERS,
            object_name=GCS_OBJ_PATH,
            bucket_name=GCS_BUCKET,
            upload_as_account=True,
            task_id="run_operator",
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute({})
        mock_ads_hook.assert_called_once_with(facebook_conn_id=FACEBOOK_ADS_CONN_ID, api_version=None)
        mock_ads_hook.return_value.bulk_facebook_report.assert_called_once_with(
            params=PARAMETERS, fields=FIELDS
        )
        mock_gcs_hook.assert_has_calls(
            [mock.call(gcp_conn_id=GCS_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)],
            [mock.call(gcp_conn_id=GCS_CONN_ID, impersonation_chain=IMPERSONATION_CHAIN)],
        )
        mock_gcs_hook.return_value.upload.assert_has_calls(
            [mock.call(bucket_name=GCS_BUCKET, object_name=GCS_OBJ_PATH_1, filename=mock.ANY, gzip=False)],
            [mock.call(bucket_name=GCS_BUCKET, object_name=GCS_OBJ_PATH_2, fidlename=mock.ANY, gzip=False)],
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Dataflow operators."""

from __future__ import annotations

import uuid
from collections.abc import Sequence
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING, Any

from googleapiclient.errors import HttpError

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.dataflow import (
    DEFAULT_DATAFLOW_LOCATION,
    DataflowHook,
)
from airflow.providers.google.cloud.links.dataflow import DataflowJobLink, DataflowPipelineLink
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.cloud.triggers.dataflow import (
    DataflowStartYamlJobTrigger,
    TemplateJobStartTrigger,
)
from airflow.providers.google.common.consts import GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID

if TYPE_CHECKING:
    from airflow.utils.context import Context


class CheckJobRunning(Enum):
    """
    Helper enum for choosing what to do if job is already running.

    IgnoreJob - do not check if running
    FinishIfRunning - finish current dag run with no action
    WaitForRun - wait for job to finish and then continue with new job
    """

    IgnoreJob = 1
    FinishIfRunning = 2
    WaitForRun = 3


class DataflowConfiguration:
    """
    Dataflow configuration for BeamRunJavaPipelineOperator and BeamRunPythonPipelineOperator.

    .. seealso::
        :class:`~airflow.providers.apache.beam.operators.beam.BeamRunJavaPipelineOperator`
        and :class:`~airflow.providers.apache.beam.operators.beam.BeamRunPythonPipelineOperator`.

    :param job_name: The 'jobName' to use when executing the Dataflow job
        (templated). This ends up being set in the pipeline options, so any entry
        with key ``'jobName'`` or  ``'job_name'``in ``options`` will be overwritten.
    :param append_job_name: True if unique suffix has to be appended to job name.
    :param project_id: Optional, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: Job location.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param poll_sleep: The time in seconds to sleep between polling Google
        Cloud Platform for the dataflow job status while the job is in the
        JOB_STATE_RUNNING state.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

        .. warning::
            This option requires Apache Beam 2.39.0 or newer.

    :param drain_pipeline: Optional, set to True if want to stop streaming job by draining it
        instead of canceling during killing task instance. See:
        https://cloud.google.com/dataflow/docs/guides/stopping-a-pipeline
    :param cancel_timeout: How long (in seconds) operator should wait for the pipeline to be
        successfully cancelled when task is being killed. (optional) default to 300s
    :param wait_until_finished: (Optional)
        If True, wait for the end of pipeline execution before exiting.
        If False, only submits job.
        If None, default behavior.

        The default behavior depends on the type of pipeline:

        * for the streaming pipeline, wait for jobs to start,
        * for the batch pipeline, wait for the jobs to complete.

        .. warning::
            You cannot call ``PipelineResult.wait_until_finish`` method in your pipeline code for the operator
            to work properly. i. e. you must use asynchronous execution. Otherwise, your pipeline will
            always wait until finished. For more information, look at:
            `Asynchronous execution
            <https://cloud.google.com/dataflow/docs/guides/specifying-exec-params#python_10>`__

        The process of starting the Dataflow job in Airflow consists of two steps:
        * running a subprocess and reading the stderr/stderr log for the job id.
        * loop waiting for the end of the job ID from the previous step by checking its status.

        Step two is started just after step one has finished, so if you have wait_until_finished in your
        pipeline code, step two will not start until the process stops. When this process stops,
        steps two will run, but it will only execute one iteration as the job will be in a terminal state.

        If you in your pipeline do not call the wait_for_pipeline method but pass wait_until_finish=True
        to the operator, the second loop will wait for the job's terminal state.

        If you in your pipeline do not call the wait_for_pipeline method, and pass wait_until_finish=False
        to the operator, the second loop will check once is job not in terminal state and exit the loop.
    :param multiple_jobs: If pipeline creates multiple jobs then monitor all jobs. Supported only by
        :class:`~airflow.providers.apache.beam.operators.beam.BeamRunJavaPipelineOperator`.
    :param check_if_running: Before running job, validate that a previous run is not in process.
        Supported only by:
        :class:`~airflow.providers.apache.beam.operators.beam.BeamRunJavaPipelineOperator`.
    :param service_account: Run the job as a specific service account, instead of the default GCE robot.
    """

    template_fields: Sequence[str] = ("job_name", "location")

    def __init__(
        self,
        *,
        job_name: str | None = None,
        append_job_name: bool = True,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str | None = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
        drain_pipeline: bool = False,
        cancel_timeout: int | None = 5 * 60,
        wait_until_finished: bool | None = None,
        multiple_jobs: bool | None = None,
        check_if_running: CheckJobRunning = CheckJobRunning.WaitForRun,
        service_account: str | None = None,
    ) -> None:
        self.job_name = job_name
        self.append_job_name = append_job_name
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain
        self.drain_pipeline = drain_pipeline
        self.cancel_timeout = cancel_timeout
        self.wait_until_finished = wait_until_finished
        self.multiple_jobs = multiple_jobs
        self.check_if_running = check_if_running
        self.service_account = service_account


class DataflowTemplatedJobStartOperator(GoogleCloudBaseOperator):
    """
    Start a Dataflow job with a classic template; the parameters of the operation will be passed to the job.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowTemplatedJobStartOperator`

    :param template: The reference to the Dataflow template.
    :param job_name: The 'jobName' to use when executing the Dataflow template
        (templated).
    :param options: Map of job runtime environment options.
        It will update environment argument if passed.

        .. seealso::
            For more information on possible configurations, look at the API documentation
            `https://cloud.google.com/dataflow/pipelines/specifying-exec-params
            <https://cloud.google.com/dataflow/docs/reference/rest/v1b3/RuntimeEnvironment>`__

    :param dataflow_default_options: Map of default job environment options.
    :param parameters: Map of job specific parameters for the template.
    :param project_id: Optional, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: Job location.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param poll_sleep: The time in seconds to sleep between polling Google
        Cloud Platform for the dataflow job status while the job is in the
        JOB_STATE_RUNNING state.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param environment: Optional, Map of job runtime environment options.

        .. seealso::
            For more information on possible configurations, look at the API documentation
            `https://cloud.google.com/dataflow/pipelines/specifying-exec-params
            <https://cloud.google.com/dataflow/docs/reference/rest/v1b3/RuntimeEnvironment>`__
    :param cancel_timeout: How long (in seconds) operator should wait for the pipeline to be
        successfully cancelled when task is being killed.
    :param append_job_name: True if unique suffix has to be appended to job name.
    :param wait_until_finished: (Optional)
        If True, wait for the end of pipeline execution before exiting.
        If False, only submits job.
        If None, default behavior.

        The default behavior depends on the type of pipeline:

        * for the streaming pipeline, wait for jobs to start,
        * for the batch pipeline, wait for the jobs to complete.

        .. warning::

            You cannot call ``PipelineResult.wait_until_finish`` method in your pipeline code for the operator
            to work properly. i. e. you must use asynchronous execution. Otherwise, your pipeline will
            always wait until finished. For more information, look at:
            `Asynchronous execution
            <https://cloud.google.com/dataflow/docs/guides/specifying-exec-params#python_10>`__

        The process of starting the Dataflow job in Airflow consists of two steps:

        * running a subprocess and reading the stderr/stderr log for the job id.
        * loop waiting for the end of the job ID from the previous step.
          This loop checks the status of the job.

        Step two is started just after step one has finished, so if you have wait_until_finished in your
        pipeline code, step two will not start until the process stops. When this process stops,
        steps two will run, but it will only execute one iteration as the job will be in a terminal state.

        If you in your pipeline do not call the wait_for_pipeline method but pass wait_until_finish=True
        to the operator, the second loop will wait for the job's terminal state.

        If you in your pipeline do not call the wait_for_pipeline method, and pass wait_until_finish=False
        to the operator, the second loop will check once is job not in terminal state and exit the loop.
    :param expected_terminal_state: The expected terminal state of the operator on which the corresponding
        Airflow task succeeds. When not specified, it will be determined by the hook.

    It's a good practice to define dataflow_* parameters in the default_args of the dag
    like the project, zone and staging location.

    .. seealso::
        https://cloud.google.com/dataflow/docs/reference/rest/v1b3/LaunchTemplateParameters
        https://cloud.google.com/dataflow/docs/reference/rest/v1b3/RuntimeEnvironment

    .. code-block:: python

       default_args = {
           "dataflow_default_options": {
               "zone": "europe-west1-d",
               "tempLocation": "gs://my-staging-bucket/staging/",
           }
       }

    You need to pass the path to your dataflow template as a file reference with the
    ``template`` parameter. Use ``parameters`` to pass on parameters to your job.
    Use ``environment`` to pass on runtime environment variables to your job.

    .. code-block:: python

       t1 = DataflowTemplatedJobStartOperator(
           task_id="dataflow_example",
           template="{{var.value.gcp_dataflow_base}}",
           parameters={
               "inputFile": "gs://bucket/input/my_input.txt",
               "outputFile": "gs://bucket/output/my_output.txt",
           },
           gcp_conn_id="airflow-conn-id",
           dag=my_dag,
       )

    ``template``, ``dataflow_default_options``, ``parameters``, and ``job_name`` are
    templated, so you can use variables in them.

    Note that ``dataflow_default_options`` is expected to save high-level options
    for project information, which apply to all dataflow operators in the DAG.

        .. seealso::
            https://cloud.google.com/dataflow/docs/reference/rest/v1b3
            /LaunchTemplateParameters
            https://cloud.google.com/dataflow/docs/reference/rest/v1b3/RuntimeEnvironment
            For more detail on job template execution have a look at the reference:
            https://cloud.google.com/dataflow/docs/templates/executing-templates

    :param deferrable: Run operator in the deferrable mode.
    """

    template_fields: Sequence[str] = (
        "template",
        "job_name",
        "options",
        "parameters",
        "project_id",
        "location",
        "gcp_conn_id",
        "impersonation_chain",
        "environment",
        "dataflow_default_options",
    )
    ui_color = "#0273d4"
    operator_extra_links = (DataflowJobLink(),)

    def __init__(
        self,
        *,
        template: str,
        project_id: str = PROVIDE_PROJECT_ID,
        job_name: str = "{{task.task_id}}",
        options: dict[str, Any] | None = None,
        dataflow_default_options: dict[str, Any] | None = None,
        parameters: dict[str, str] | None = None,
        location: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
        environment: dict | None = None,
        cancel_timeout: int | None = 10 * 60,
        wait_until_finished: bool | None = None,
        append_job_name: bool = True,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        expected_terminal_state: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.template = template
        self.job_name = job_name
        self.options = options or {}
        self.dataflow_default_options = dataflow_default_options or {}
        self.parameters = parameters or {}
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.poll_sleep = poll_sleep
        self.impersonation_chain = impersonation_chain
        self.environment = environment
        self.cancel_timeout = cancel_timeout
        self.wait_until_finished = wait_until_finished
        self.append_job_name = append_job_name
        self.deferrable = deferrable
        self.expected_terminal_state = expected_terminal_state

        self.job: dict[str, str] | None = None

        self._validate_deferrable_params()

    def _validate_deferrable_params(self):
        if self.deferrable and self.wait_until_finished:
            raise ValueError(
                "Conflict between deferrable and wait_until_finished parameters "
                "because it makes operator as blocking when it requires to be deferred. "
                "It should be True as deferrable parameter or True as wait_until_finished."
            )

        if self.deferrable and self.wait_until_finished is None:
            self.wait_until_finished = False

    @cached_property
    def hook(self) -> DataflowHook:
        hook = DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
            cancel_timeout=self.cancel_timeout,
            wait_until_finished=self.wait_until_finished,
            expected_terminal_state=self.expected_terminal_state,
        )
        return hook

    def execute(self, context: Context):
        def set_current_job(current_job):
            self.job = current_job
            DataflowJobLink.persist(
                context=context,
                project_id=self.project_id,
                region=self.location,
                job_id=self.job.get("id"),
            )

        options = self.dataflow_default_options
        options.update(self.options)

        if not self.location:
            self.location = DEFAULT_DATAFLOW_LOCATION

        if not self.deferrable:
            self.job = self.hook.start_template_dataflow(
                job_name=self.job_name,
                variables=options,
                parameters=self.parameters,
                dataflow_template=self.template,
                on_new_job_callback=set_current_job,
                project_id=self.project_id,
                location=self.location,
                environment=self.environment,
                append_job_name=self.append_job_name,
            )
            job_id = self.hook.extract_job_id(self.job)
            context["task_instance"].xcom_push(key="job_id", value=job_id)
            return job_id

        self.job = self.hook.launch_job_with_template(
            job_name=self.job_name,
            variables=options,
            parameters=self.parameters,
            dataflow_template=self.template,
            project_id=self.project_id,
            append_job_name=self.append_job_name,
            location=self.location,
            environment=self.environment,
        )
        job_id = self.hook.extract_job_id(self.job)
        DataflowJobLink.persist(
            context=context, project_id=self.project_id, region=self.location, job_id=job_id
        )
        self.defer(
            trigger=TemplateJobStartTrigger(
                project_id=self.project_id,
                job_id=job_id,
                location=self.location,
                gcp_conn_id=self.gcp_conn_id,
                poll_sleep=self.poll_sleep,
                impersonation_chain=self.impersonation_chain,
                cancel_timeout=self.cancel_timeout,
            ),
            method_name=GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
        )

    def execute_complete(self, context: Context, event: dict[str, Any]) -> str:
        """Execute after trigger finishes its work."""
        if event["status"] in ("error", "stopped"):
            self.log.info("status: %s, msg: %s", event["status"], event["message"])
            raise AirflowException(event["message"])

        job_id = event["job_id"]
        context["task_instance"].xcom_push(key="job_id", value=job_id)
        self.log.info("Task %s completed with response %s", self.task_id, event["message"])
        return job_id

    def on_kill(self) -> None:
        self.log.info("On kill.")
        if self.job is not None:
            self.log.info("Cancelling job %s", self.job_name)
            self.hook.cancel_job(
                job_name=self.job_name,
                job_id=self.job.get("id"),
                project_id=self.job.get("projectId"),
                location=self.job.get("location"),
            )


class DataflowStartFlexTemplateOperator(GoogleCloudBaseOperator):
    """
    Starts a Dataflow Job with a Flex Template.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowStartFlexTemplateOperator`

    :param body: The request body. See:
        https://cloud.google.com/dataflow/docs/reference/rest/v1b3/projects.locations.flexTemplates/launch#request-body
    :param location: The location of the Dataflow job (for example europe-west1)
    :param project_id: The ID of the GCP project that owns the job.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud
        Platform.
    :param drain_pipeline: Optional, set to True if want to stop streaming job by draining it
        instead of canceling during killing task instance. See:
        https://cloud.google.com/dataflow/docs/guides/stopping-a-pipeline
    :param cancel_timeout: How long (in seconds) operator should wait for the pipeline to be
        successfully cancelled when task is being killed.
    :param wait_until_finished: (Optional)
        If True, wait for the end of pipeline execution before exiting.
        If False, only submits job.
        If None, default behavior.

        The default behavior depends on the type of pipeline:

        * for the streaming pipeline, wait for jobs to start,
        * for the batch pipeline, wait for the jobs to complete.

        .. warning::

            You cannot call ``PipelineResult.wait_until_finish`` method in your pipeline code for the operator
            to work properly. i. e. you must use asynchronous execution. Otherwise, your pipeline will
            always wait until finished. For more information, look at:
            `Asynchronous execution
            <https://cloud.google.com/dataflow/docs/guides/specifying-exec-params#python_10>`__

        The process of starting the Dataflow job in Airflow consists of two steps:

        * running a subprocess and reading the stderr/stderr log for the job id.
        * loop waiting for the end of the job ID from the previous step.
          This loop checks the status of the job.

        Step two is started just after step one has finished, so if you have wait_until_finished in your
        pipeline code, step two will not start until the process stops. When this process stops,
        steps two will run, but it will only execute one iteration as the job will be in a terminal state.

        If you in your pipeline do not call the wait_for_pipeline method but pass wait_until_finished=True
        to the operator, the second loop will wait for the job's terminal state.

        If you in your pipeline do not call the wait_for_pipeline method, and pass wait_until_finished=False
        to the operator, the second loop will check once is job not in terminal state and exit the loop.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: Run operator in the deferrable mode.
    :param expected_terminal_state: The expected final status of the operator on which the corresponding
        Airflow task succeeds. When not specified, it will be determined by the hook.
    :param append_job_name: True if unique suffix has to be appended to job name.
    :param poll_sleep: The time in seconds to sleep between polling Google
        Cloud Platform for the dataflow job status while the job is in the
        JOB_STATE_RUNNING state.
    """

    template_fields: Sequence[str] = ("body", "location", "project_id", "gcp_conn_id")
    operator_extra_links = (DataflowJobLink(),)

    def __init__(
        self,
        body: dict,
        location: str,
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        drain_pipeline: bool = False,
        cancel_timeout: int | None = 10 * 60,
        wait_until_finished: bool | None = None,
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        append_job_name: bool = True,
        expected_terminal_state: str | None = None,
        poll_sleep: int = 10,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.body = body
        self.location = location
        self.project_id = project_id
        self.gcp_conn_id = gcp_conn_id
        self.drain_pipeline = drain_pipeline
        self.cancel_timeout = cancel_timeout
        self.wait_until_finished = wait_until_finished
        self.job: dict[str, str] | None = None
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.expected_terminal_state = expected_terminal_state
        self.append_job_name = append_job_name
        self.poll_sleep = poll_sleep

        self._validate_deferrable_params()

    def _validate_deferrable_params(self):
        if self.deferrable and self.wait_until_finished:
            raise ValueError(
                "Conflict between deferrable and wait_until_finished parameters "
                "because it makes operator as blocking when it requires to be deferred. "
                "It should be True as deferrable parameter or True as wait_until_finished."
            )

        if self.deferrable and self.wait_until_finished is None:
            self.wait_until_finished = False

    @cached_property
    def hook(self) -> DataflowHook:
        hook = DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            drain_pipeline=self.drain_pipeline,
            cancel_timeout=self.cancel_timeout,
            wait_until_finished=self.wait_until_finished,
            impersonation_chain=self.impersonation_chain,
            expected_terminal_state=self.expected_terminal_state,
        )
        return hook

    def execute(self, context: Context):
        if self.append_job_name:
            self._append_uuid_to_job_name()

        def set_current_job(current_job):
            self.job = current_job
            DataflowJobLink.persist(
                context=context, project_id=self.project_id, region=self.location, job_id=self.job.get("id")
            )

        if not self.deferrable:
            self.job = self.hook.start_flex_template(
                body=self.body,
                location=self.location,
                project_id=self.project_id,
                on_new_job_callback=set_current_job,
            )
            job_id = self.hook.extract_job_id(self.job)
            context["task_instance"].xcom_push(key="job_id", value=job_id)
            return self.job

        self.job = self.hook.launch_job_with_flex_template(
            body=self.body,
            location=self.location,
            project_id=self.project_id,
        )
        job_id = self.hook.extract_job_id(self.job)
        DataflowJobLink.persist(
            context=context, project_id=self.project_id, region=self.location, job_id=job_id
        )
        self.defer(
            trigger=TemplateJobStartTrigger(
                project_id=self.project_id,
                job_id=job_id,
                location=self.location,
                gcp_conn_id=self.gcp_conn_id,
                poll_sleep=self.poll_sleep,
                impersonation_chain=self.impersonation_chain,
                cancel_timeout=self.cancel_timeout,
            ),
            method_name=GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
        )

    def _append_uuid_to_job_name(self):
        job_body = self.body.get("launch_parameter") or self.body.get("launchParameter")
        job_name = job_body.get("jobName")
        if job_name:
            job_name += f"-{uuid.uuid4()!s:.8}"
            job_body["jobName"] = job_name
            self.log.info("Job name was changed to %s", job_name)

    def execute_complete(self, context: Context, event: dict) -> dict[str, str]:
        """Execute after trigger finishes its work."""
        if event["status"] in ("error", "stopped"):
            self.log.info("status: %s, msg: %s", event["status"], event["message"])
            raise AirflowException(event["message"])

        job_id = event["job_id"]
        self.log.info("Task %s completed with response %s", job_id, event["message"])
        context["task_instance"].xcom_push(key="job_id", value=job_id)
        job = self.hook.get_job(job_id=job_id, project_id=self.project_id, location=self.location)
        return job

    def on_kill(self) -> None:
        self.log.info("On kill.")
        if self.job is not None:
            self.hook.cancel_job(
                job_id=self.job.get("id"),
                project_id=self.job.get("projectId"),
                location=self.job.get("location"),
            )


class DataflowStartYamlJobOperator(GoogleCloudBaseOperator):
    """
    Launch a Dataflow YAML job and return the result.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowStartYamlJobOperator`

    .. warning::
        This operator requires ``gcloud`` command (Google Cloud SDK) must be installed on the Airflow worker
        <https://cloud.google.com/sdk/docs/install>`__

    :param job_name: Required. The unique name to assign to the Cloud Dataflow job.
    :param yaml_pipeline_file: Required. Path to a file defining the YAML pipeline to run.
        Must be a local file or a URL beginning with 'gs://'.
    :param region: Optional. Region ID of the job's regional endpoint. Defaults to 'us-central1'.
    :param project_id: Required. The ID of the GCP project that owns the job.
        If set to ``None`` or missing, the default project_id from the GCP connection is used.
    :param gcp_conn_id: Optional. The connection ID used to connect to GCP.
    :param append_job_name: Optional. Set to True if a unique suffix has to be appended to the `job_name`.
        Defaults to True.
    :param drain_pipeline: Optional. Set to True if you want to stop a streaming pipeline job by draining it
        instead of canceling when killing the task instance. Note that this does not work for batch pipeline jobs
        or in the deferrable mode. Defaults to False.
        For more info see: https://cloud.google.com/dataflow/docs/guides/stopping-a-pipeline
    :param deferrable: Optional. Run operator in the deferrable mode.
    :param expected_terminal_state: Optional. The expected terminal state of the Dataflow job at which the
        operator task is set to succeed. Defaults to 'JOB_STATE_DONE' for the batch jobs and 'JOB_STATE_RUNNING'
        for the streaming jobs.
    :param poll_sleep: Optional. The time in seconds to sleep between polling Google Cloud Platform for the Dataflow job status.
        Used both for the sync and deferrable mode.
    :param cancel_timeout: Optional. How long (in seconds) operator should wait for the pipeline to be
        successfully canceled when the task is being killed.
    :param jinja_variables: Optional. A dictionary of Jinja2 variables to be used in reifying the yaml pipeline file.
    :param options: Optional. Additional gcloud or Beam job parameters.
        It must be a dictionary with the keys matching the optional flag names in gcloud.
        The list of supported flags can be found at: `https://cloud.google.com/sdk/gcloud/reference/dataflow/yaml/run`.
        Note that if a flag does not require a value, then its dictionary value must be either True or None.
        For example, the `--log-http` flag can be passed as {'log-http': True}.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :return: Dictionary containing the job's data.
    """

    template_fields: Sequence[str] = (
        "job_name",
        "yaml_pipeline_file",
        "jinja_variables",
        "options",
        "region",
        "project_id",
        "gcp_conn_id",
    )
    template_fields_renderers = {
        "jinja_variables": "json",
    }
    operator_extra_links = (DataflowJobLink(),)

    def __init__(
        self,
        *,
        job_name: str,
        yaml_pipeline_file: str,
        region: str = DEFAULT_DATAFLOW_LOCATION,
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        append_job_name: bool = True,
        drain_pipeline: bool = False,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        poll_sleep: int = 10,
        cancel_timeout: int | None = 5 * 60,
        expected_terminal_state: str | None = None,
        jinja_variables: dict[str, str] | None = None,
        options: dict[str, Any] | None = None,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.job_name = job_name
        self.yaml_pipeline_file = yaml_pipeline_file
        self.region = region
        self.project_id = project_id
        self.gcp_conn_id = gcp_conn_id
        self.append_job_name = append_job_name
        self.drain_pipeline = drain_pipeline
        self.deferrable = deferrable
        self.poll_sleep = poll_sleep
        self.cancel_timeout = cancel_timeout
        self.expected_terminal_state = expected_terminal_state
        self.options = options
        self.jinja_variables = jinja_variables
        self.impersonation_chain = impersonation_chain
        self.job_id: str | None = None

    def execute(self, context: Context) -> dict[str, Any]:
        self.job_id = self.hook.launch_beam_yaml_job(
            job_name=self.job_name,
            yaml_pipeline_file=self.yaml_pipeline_file,
            append_job_name=self.append_job_name,
            options=self.options,
            jinja_variables=self.jinja_variables,
            project_id=self.project_id,
            location=self.region,
        )

        DataflowJobLink.persist(
            context=context, project_id=self.project_id, region=self.region, job_id=self.job_id
        )

        if self.deferrable:
            self.defer(
                trigger=DataflowStartYamlJobTrigger(
                    job_id=self.job_id,
                    project_id=self.project_id,
                    location=self.region,
                    gcp_conn_id=self.gcp_conn_id,
                    poll_sleep=self.poll_sleep,
                    cancel_timeout=self.cancel_timeout,
                    expected_terminal_state=self.expected_terminal_state,
                    impersonation_chain=self.impersonation_chain,
                ),
                method_name=GOOGLE_DEFAULT_DEFERRABLE_METHOD_NAME,
            )

        self.hook.wait_for_done(
            job_name=self.job_name, location=self.region, project_id=self.project_id, job_id=self.job_id
        )
        job = self.hook.get_job(job_id=self.job_id, location=self.region, project_id=self.project_id)
        return job

    def execute_complete(self, context: Context, event: dict) -> dict[str, Any]:
        """Execute after the trigger returns an event."""
        if event["status"] in ("error", "stopped"):
            self.log.info("status: %s, msg: %s", event["status"], event["message"])
            raise AirflowException(event["message"])
        job = event["job"]
        self.log.info("Job %s completed with response %s", job["id"], event["message"])
        context["task_instance"].xcom_push(key="job_id", value=job["id"])

        return job

    def on_kill(self):
        """
        Cancel the dataflow job if a task instance gets killed.

        This method will not be called if a task instance is killed in a deferred
        state.
        """
        self.log.info("On kill called.")
        if self.job_id:
            self.hook.cancel_job(
                job_id=self.job_id,
                project_id=self.project_id,
                location=self.region,
            )

    @cached_property
    def hook(self) -> DataflowHook:
        return DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
            drain_pipeline=self.drain_pipeline,
            cancel_timeout=self.cancel_timeout,
            expected_terminal_state=self.expected_terminal_state,
        )


class DataflowStopJobOperator(GoogleCloudBaseOperator):
    """
    Stops the job with the specified name prefix or Job ID.

    All jobs with provided name prefix will be stopped.
    Streaming jobs are drained by default.

    Parameter ``job_name_prefix`` and ``job_id`` are mutually exclusive.

    .. seealso::
        For more details on stopping a pipeline see:
        https://cloud.google.com/dataflow/docs/guides/stopping-a-pipeline

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowStopJobOperator`

    :param job_name_prefix: Name prefix specifying which jobs are to be stopped.
    :param job_id: Job ID specifying which jobs are to be stopped.
    :param project_id: Optional, the Google Cloud project ID in which to start a job.
        If set to None or missing, the default project_id from the Google Cloud connection is used.
    :param location: Optional, Job location. If set to None or missing, "us-central1" will be used.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param poll_sleep: The time in seconds to sleep between polling Google
        Cloud Platform for the dataflow job status to confirm it's stopped.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param drain_pipeline: Optional, set to False if want to stop streaming job by canceling it
        instead of draining. See: https://cloud.google.com/dataflow/docs/guides/stopping-a-pipeline
    :param stop_timeout: wait time in seconds for successful job canceling/draining
    """

    template_fields = [
        "job_id",
        "project_id",
        "impersonation_chain",
    ]

    def __init__(
        self,
        job_name_prefix: str | None = None,
        job_id: str | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        poll_sleep: int = 10,
        impersonation_chain: str | Sequence[str] | None = None,
        stop_timeout: int | None = 10 * 60,
        drain_pipeline: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.poll_sleep = poll_sleep
        self.stop_timeout = stop_timeout
        self.job_name = job_name_prefix
        self.job_id = job_id
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.hook: DataflowHook | None = None
        self.drain_pipeline = drain_pipeline

    def execute(self, context: Context) -> None:
        self.dataflow_hook = DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            poll_sleep=self.poll_sleep,
            impersonation_chain=self.impersonation_chain,
            cancel_timeout=self.stop_timeout,
            drain_pipeline=self.drain_pipeline,
        )
        if self.job_id or self.dataflow_hook.is_job_dataflow_running(
            name=self.job_name,
            project_id=self.project_id,
            location=self.location,
        ):
            self.dataflow_hook.cancel_job(
                job_name=self.job_name,
                project_id=self.project_id,
                location=self.location,
                job_id=self.job_id,
            )
        else:
            self.log.info("No jobs to stop")

        return None


class DataflowCreatePipelineOperator(GoogleCloudBaseOperator):
    """
    Creates a new Dataflow Data Pipeline instance.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowCreatePipelineOperator`

    :param body: The request body (contains instance of Pipeline). See:
        https://cloud.google.com/dataflow/docs/reference/data-pipelines/rest/v1/projects.locations.pipelines/create#request-body
    :param project_id: The ID of the GCP project that owns the job.
    :param location: The location to direct the Data Pipelines instance to (for example us-central1).
    :param gcp_conn_id: The connection ID to connect to the Google Cloud
        Platform.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

        .. warning::
            This option requires Apache Beam 2.39.0 or newer.

    Returns the created Dataflow Data Pipeline instance in JSON representation.
    """

    operator_extra_links = (DataflowPipelineLink(),)

    def __init__(
        self,
        *,
        body: dict,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.body = body
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.dataflow_hook: DataflowHook | None = None

        self.pipeline_name = self.body["name"].split("/")[-1] if self.body else None

    @property
    def extra_links_params(self) -> dict[str, Any]:
        return {
            "project_id": self.project_id,
            "location": self.location,
            "pipeline_name": self.pipeline_name,
        }

    def execute(self, context: Context):
        if self.body is None:
            raise AirflowException(
                "Request Body not given; cannot create a Data Pipeline without the Request Body."
            )
        if self.project_id is None:
            raise AirflowException(
                "Project ID not given; cannot create a Data Pipeline without the Project ID."
            )
        if self.location is None:
            raise AirflowException("location not given; cannot create a Data Pipeline without the location.")

        self.dataflow_hook = DataflowHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        self.body["pipelineSources"] = {"airflow": "airflow"}
        try:
            self.pipeline = self.dataflow_hook.create_data_pipeline(
                project_id=self.project_id,
                body=self.body,
                location=self.location,
            )
        except HttpError as e:
            if e.resp.status == 409:
                # If the pipeline already exists, retrieve it
                self.log.info("Pipeline with given name already exists.")
                self.pipeline = self.dataflow_hook.get_data_pipeline(
                    project_id=self.project_id,
                    pipeline_name=self.pipeline_name,
                    location=self.location,
                )
        DataflowPipelineLink.persist(context=context)
        context["task_instance"].xcom_push(key="pipeline_name", value=self.pipeline_name)
        if self.pipeline:
            if "error" in self.pipeline:
                raise AirflowException(self.pipeline.get("error").get("message"))

        return self.pipeline


class DataflowRunPipelineOperator(GoogleCloudBaseOperator):
    """
    Runs a Dataflow Data Pipeline.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowRunPipelineOperator`

    :param pipeline_name:  The display name of the pipeline. In example
        projects/PROJECT_ID/locations/LOCATION_ID/pipelines/PIPELINE_ID it would be the PIPELINE_ID.
    :param project_id: The ID of the GCP project that owns the job.
    :param location: The location to direct the Data Pipelines instance to (for example us-central1).
    :param gcp_conn_id: The connection ID to connect to the Google Cloud Platform.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

    Returns the created Job in JSON representation.
    """

    operator_extra_links = (DataflowJobLink(),)

    def __init__(
        self,
        pipeline_name: str,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.pipeline_name = pipeline_name
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.dataflow_hook: DataflowHook | None = None

    def execute(self, context: Context):
        self.dataflow_hook = DataflowHook(
            gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain
        )

        if self.pipeline_name is None:
            raise AirflowException("Data Pipeline name not given; cannot run unspecified pipeline.")
        if self.project_id is None:
            raise AirflowException("Data Pipeline Project ID not given; cannot run pipeline.")
        if self.location is None:
            raise AirflowException("Data Pipeline location not given; cannot run pipeline.")
        try:
            self.job = self.dataflow_hook.run_data_pipeline(
                pipeline_name=self.pipeline_name,
                project_id=self.project_id,
                location=self.location,
            )["job"]
            job_id = self.dataflow_hook.extract_job_id(self.job)
            context["task_instance"].xcom_push(key="job_id", value=job_id)
            DataflowJobLink.persist(
                context=context, project_id=self.project_id, region=self.location, job_id=job_id
            )
        except HttpError as e:
            if e.resp.status == 404:
                raise AirflowException("Pipeline with given name was not found.")
        except Exception as exc:
            raise AirflowException("Error occurred when running Pipeline: %s", exc)

        return self.job


class DataflowDeletePipelineOperator(GoogleCloudBaseOperator):
    """
    Deletes a Dataflow Data Pipeline.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DataflowDeletePipelineOperator`

    :param pipeline_name: The display name of the pipeline. In example
        projects/PROJECT_ID/locations/LOCATION_ID/pipelines/PIPELINE_ID it would be the PIPELINE_ID.
    :param project_id: The ID of the GCP project that owns the job.
    :param location: The location to direct the Data Pipelines instance to (for example us-central1).
    :param gcp_conn_id: The connection ID to connect to the Google Cloud Platform.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    def __init__(
        self,
        pipeline_name: str,
        project_id: str = PROVIDE_PROJECT_ID,
        location: str = DEFAULT_DATAFLOW_LOCATION,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.pipeline_name = pipeline_name
        self.project_id = project_id
        self.location = location
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.dataflow_hook: DataflowHook | None = None
        self.response: dict | None = None

    def execute(self, context: Context):
        self.dataflow_hook = DataflowHook(
            gcp_conn_id=self.gcp_conn_id, impersonation_chain=self.impersonation_chain
        )

        if self.pipeline_name is None:
            raise AirflowException("Data Pipeline name not given; cannot run unspecified pipeline.")
        if self.project_id is None:
            raise AirflowException("Data Pipeline Project ID not given; cannot run pipeline.")
        if self.location is None:
            raise AirflowException("Data Pipeline location not given; cannot run pipeline.")

        self.response = self.dataflow_hook.delete_data_pipeline(
            pipeline_name=self.pipeline_name,
            project_id=self.project_id,
            location=self.location,
        )

        if self.response:
            raise AirflowException(self.response)

        return None
